/*
 *  i2c_pxa.c
 *
 *  I2C adapter for the PXA I2C bus access
 *
 *  Copyright (C) 2002 Intrinsyc Software Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  History:
 *    Apr 2002: Initial version [CS]
 *    Jun 2002: Properly seperated algo/adap [FB]
 *    Jan 2003: Fixed several bugs concerning interrupt handling [Kai-Uwe Bloem]
 *    Jan 2003: added limited signal handling [Kai-Uwe Bloem]
 *    Jun 2003: updated for 2.5 [Dustin McIntire]
 */

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/interrupt.h>

#include <asm/arch/pxa-regs.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/arch/irqs.h>	      /* for IRQ_I2C */
#include <asm/arch/i2c-pxa.h>

/*
 * Set this to zero to remove all debug statements via dead code elimination.
 */
//#define DEBUG       1

#if DEBUG
static unsigned int i2c_debug = DEBUG;
#else
#define i2c_debug	0
#endif

static int irq = 0;
static volatile int i2c_pending = 0;	     /* interrupt pending when 1 */
static volatile int bus_error = 0;
static volatile int tx_finished = 0;
static volatile int rx_finished = 0;
static volatile int do_slave_receive = 0;
static volatile char srec_buffer[32];
static volatile int srec_count=0;


static wait_queue_head_t i2c_wait;
static void i2c_pxa_transfer( int lastbyte, int receive, int midbyte);

static spinlock_t i2c_pxa_irqlock = SPIN_LOCK_UNLOCKED;

DECLARE_WAIT_QUEUE_HEAD(slave_receive_queue);

/* Register slave receive callback function   */
/* Should do a linked list here, but I'm lazy */

int i2c_pxa_slave_recv(char *buffer, int size)
{
//  int i;
  sleep_on(&slave_receive_queue);
  memcpy(buffer, &srec_buffer, srec_count);
  
//  printk("i2c-pxa.ko: ");
//  for(i=0;i<srec_count;i++)
//    printk("%d ", buffer[i]);
//  printk("\n");
  
  return srec_count;
}


/* place a byte in the transmit register */
static void i2c_pxa_write_byte(u8 value)
{
	IDBR = value;
}

/* read byte in the receive register */
static u8 i2c_pxa_read_byte(void)
{
	return (u8) (0xff & IDBR);
}

static void i2c_pxa_start(void)
{
	unsigned long icr = ICR;
	icr |= ICR_START;
	icr &= ~(ICR_STOP | ICR_ALDIE | ICR_ACKNAK);
	ICR = icr;

	bus_error=0;	    /* clear any bus_error from previous txfers */
	tx_finished=0;	  /* clear rx and tx interrupts from previous txfers */
	rx_finished=0;
	i2c_pending = 0;
}

static void i2c_pxa_repeat_start(void)
{
	unsigned long icr = ICR;
	icr |= ICR_START;
	icr &= ~(ICR_STOP | ICR_ALDIE);
	ICR = icr;

	bus_error=0;	    /* clear any bus_error from previous txfers */
	tx_finished=0;	  /* clear rx and tx interrupts from previous txfers */
	rx_finished=0;
	i2c_pending = 0;
}

static void i2c_pxa_stop(void)
{
	unsigned long icr = ICR;

    /* Intel PXA255 Processor Developer's Manual */
	/* Page 9-19, do step # 7                    */
	icr |= ICR_STOP;
	icr &= ~(ICR_START);
	ICR = icr;

}

static void i2c_pxa_midbyte(void)
{
	unsigned long icr = ICR;
	icr &= ~(ICR_START | ICR_STOP);
	ICR = icr;
}

static void i2c_pxa_abort(void)
{
	unsigned long timeout = jiffies + HZ/4;

#ifdef PXA_ABORT_MA
	while ((long)(timeout - jiffies) > 0 && (ICR & ICR_TB)) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(1);
	}

	ICR |= ICR_MA;
	udelay(100);
#else
	while ((long)(timeout - jiffies) > 0 && (IBMR & 0x1) == 0) {
		i2c_pxa_transfer( 1, I2C_RECEIVE, 1);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(1);
	}
#endif
	ICR &= ~(ICR_MA | ICR_START | ICR_STOP);
}

static int i2c_pxa_wait_bus_not_busy(void)
{
	int timeout = DEF_TIMEOUT;

	while (timeout-- && (ISR & ISR_IBB)) {
		udelay(100); /* wait for 100 us */
	}

	return (timeout<=0);
}

static void i2c_pxa_wait_for_ite(void)
{ 
	unsigned long flags;
	if (irq > 0) {
	    spin_lock_irqsave(&i2c_pxa_irqlock, flags);
		if (i2c_pending == 0) {
			interruptible_sleep_on_timeout(&i2c_wait, I2C_SLEEP_TIMEOUT );
		}
		i2c_pending = 0;
		spin_unlock_irqrestore(&i2c_pxa_irqlock, flags);
	} else {
		udelay(100);
	}
}

static int i2c_pxa_wait_for_int(int wait_type)
{
	int timeout = DEF_TIMEOUT;
#ifdef DEBUG
	if (bus_error)
		printk(KERN_INFO"i2c_pxa_wait_for_int: Bus error on enter\n");
	if (rx_finished)
		printk(KERN_INFO"i2c_pxa_wait_for_int: Receive interrupt on enter\n");
	if (tx_finished)
		printk(KERN_INFO"i2c_pxa_wait_for_int: Transmit interrupt on enter\n");
#endif

	if (wait_type == I2C_RECEIVE) {	/* wait on receive */

		do {
			i2c_pxa_wait_for_ite();
		} while (!(rx_finished) && timeout-- && !signal_pending(current));

#ifdef DEBUG
		if (timeout<0) {
			if (tx_finished)
				printk("I2C: received a tx interrupt while waiting on a rx in wait_for_int");
		}
#endif
	} else {		  /* wait on transmit */

		do {
			i2c_pxa_wait_for_ite();
		} while (!(tx_finished) && timeout-- && !signal_pending(current));

#ifdef DEBUG
		if (timeout<0) {
			if (rx_finished)
				printk("i2c: received a rx interrupt while waiting on a tx in wait_for_int");
		}
#endif
	}

	udelay(ACK_DELAY);      /* this is needed for the bus error */

	tx_finished=0;
	rx_finished=0;

	if (bus_error) {
		bus_error=0;
		if (i2c_debug > 2)printk("wait_for_int: error - no ack\n");
		return BUS_ERROR;
	}

	if (signal_pending(current)) {
		return (-ERESTARTSYS);
	} else
	if (timeout < 0) {
		if (i2c_debug > 2)printk("wait_for_int: timeout\n");
		return(-EIO);
	} else
		return(0);
}

static void i2c_pxa_transfer(int lastbyte, int receive, int midbyte)
{
	if (lastbyte) {
/*
		if (receive==I2C_RECEIVE)
		{
		    printk("about to set ICR_ACKNAK\n"); 
		    ICR |= ICR_ACKNAK;
		}
*/
		i2c_pxa_stop();  /* pxa doc, page 9-19 step 7 */		
	} else
	if (midbyte) {
		i2c_pxa_midbyte();
	}
	ICR |= ICR_TB;
}

static void i2c_pxa_reset( void)
{
#ifdef DEBUG
	printk("Resetting I2C Controller Unit\n");
#endif

	/* abort any transfer currently under way */
	i2c_pxa_abort();

	/* reset according to 9.8 */
	ICR = ICR_UR;
	ISR = I2C_ISR_INIT;
	
	
	ICR &= ~ICR_UR;

	/* set the global I2C clock on */
	CKEN |= CKEN14_I2C;

	/* set our slave address */
	ISAR = I2C_PXA_SLAVE_ADDR;
    //ISAR = 0x06;
	/* set control register values */
	ICR = I2C_ICR_INIT;

	ICR |= ICR_SADIE | ICR_SSDIE | 0x00008000;

	/* clear any leftover states from prior transmissions */
	i2c_pending = rx_finished = tx_finished = bus_error = 0;

	/* enable unit */
	ICR |= ICR_IUE;
	udelay(100);
}

static irqreturn_t i2c_pxa_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	int config, status, wakeup = 0;
	int do_wakeup = 1;

	config=(ICR);
	status=(ISR);

#ifdef BRIANDEBUG
    printk("----------------------------------\n");
	printk("jiffies: %lu\n", jiffies);
	printk("ICR = 0x%8.8X\n", config);
	if((config) & 0x00000001)
	  printk("ICR_START ");
	if((config) & 0x00000002)
	  printk("ICR_STOP ");
	if((config) & 0x00000004)
	  printk("ICR_ACKNAK ");
    if((config) & 0x00000008)
	  printk("ICR_TB ");
	if((config) & 0x00000010)
	  printk("ICR_MA ");
	if((config) & 0x00000020)
	  printk("ICR_SCLE ");
	if((config) & 0x00000040)
	  printk("ICR_IUE ");
	if((config) & 0x00000080)
	  printk("ICR_GCD ");
	if((config) & 0x00000100)
	  printk("ICR_ITEIE ");
	if((config) & 0x00000200)
	  printk("ICR_IRFIE ");
	if((config) & 0x00000400)
	  printk("ICR_BEIE ");
	if((config) & 0x00000800)
	  printk("ICR_SSDIE ");
	if((config) & 0x00001000)
	  printk("ICR_ALDIE ");
	if((config) & 0x00002000)
	  printk("ICR_SADIE ");
	if((config) & 0x00004000)
	  printk("ICR_UR ");
	if((config) & 0x00008000)
	  printk("ICR_FM ");
	printk("\n");
	
	printk("ISR = 0x%8.8X\n", status);
	if((status) & ISR_RWM)
	  printk("ISR_RWM ");
	if((status) & ISR_ACKNAK)
	  printk("ISR_ACKNAK ");
	if((status) & ISR_UB)
	  printk("ISR_UB ");
	if((status) & ISR_IBB)
	  printk("ISR_IBB ");
	if((status) & ISR_SSD)
	  printk("ISR_SSD ");
	if((status) & ISR_ALD)
	  printk("ISR_ALD ");
	if((status) & ISR_ITE)
	  printk("ISR_ITE ");
	if((status) & 0x00000080)
	  printk("ISR_IRF ");
	if((status) & 0x00000100)
	  printk("ISR_GCAD ");
	if((status) & 0x00000200)
	  printk("ISR_SAD ");
	if((status) & 0x00000400)
	  printk("ISR_BED ");
	printk("\n");
	
	
	printk("IDBR: %d\n", (IDBR));
#endif

    if ((status & ISR_ITE) && (config & ICR_STOP) && (!(status & ISR_RWM)))
	{
#ifdef BRIANDEBUG
	  printk("should do step 9 and 10\n");
#endif

	  /* pxa doc, page 9-19 - step 9 */
	  ISR |= ISR_ITE;
	  	  
	  /* pxa doc, page 9-19 - step 10 */
	  ICR = config & ~(ICR_STOP);


	  do_wakeup = 0;
	  //return IRQ_HANDLED;
	}

    if (status & ISR_ALD)
	{
 	    printk("Arbitration lost\n");
		ISR |= ISR_ALD;
	}
	if ((status & ISR_SAD) && (status & ISR_UB) && !(status & ISR_RWM))// && (!(status & ISR_ITE)))
	{
#ifdef BRIANDEBUG
		printk("%u i2c-pxa: Slave Address Detected\n", jiffies);
#endif
        srec_count = 0;
		do_slave_receive = 1;
//		PREPARE_WORK(&i2c_pxa_work, (void *)i2c_pxa_slave_receive, NULL);
//		schedule_work(&i2c_pxa_work);
           
		(ISR) |= ISR_SAD;   /* acknowledge that our address has been called */
		(ICR) |= ICR_TB;	/* go ahead and send us the next byte */
		return IRQ_HANDLED;
	}

    if ((status & ISR_SSD) && do_slave_receive)
	{
//	  int i;
	  
//	  printk("srec_buffer: ");
//	  for(i=0; i< srec_count; i++)
//	    printk("%d ", srec_buffer[i]);
//	  printk("\n");
//	  printk("i2c_pxa: wake_up(&slave_receive_queue\n");
	  wake_up(&slave_receive_queue);
/*
	  if(srec_callback != NULL)
	  {
	    srec_callback(srec_count, srec_buffer);
	  }
*/
#ifdef BRIANDEBUG
	    printk("Slave Received the following message: ");
	    for(i=0; i< srec_count; i++)
 	      printk("%d ", srec_buffer[i]); 
	    printk("\n");
#endif
#ifdef BRIANDEBUG
	  printk("\nend of slave receive\n");
#endif
	  (ICR) |= ICR_TB;
	  (ISR) |= ISR_SSD;
	  return IRQ_HANDLED;
	}	
	if (status & ISR_SSD)
	{
	  printk("clearing ISR_SSD\n");
	  (ISR) |= ISR_SSD;
	  return IRQ_HANDLED;
	}
	if (status & ISR_SAD)
	{
	  printk("clearning ISR_SAD (not a slave receive)\n");
	  (ISR) |= ISR_SAD;
	  return IRQ_HANDLED;
	}

    if ((status & ISR_IRF) && do_slave_receive)
	{
	   srec_buffer[srec_count] = (IDBR);
	   srec_count++;
#ifdef BRIANDEBUG
	   printk("next byte: %d\n", (IDBR));
#endif
	   (ISR) |= ISR_IRF;
	   (ICR) |= ICR_TB;
	   return IRQ_HANDLED;
	}

	if (status & ISR_BED) {
		(ISR) |= ISR_BED;
		bus_error=ISR_BED;
		wakeup = 1;
	}
	if (status & ISR_ITE) {
		(ISR) |= ISR_ITE;
		tx_finished=ISR_ITE;
		wakeup = 1;
	}
	if (status & ISR_IRF) {
		(ISR) |= ISR_IRF;
		rx_finished=ISR_IRF;
		wakeup = 1;
	}
	if (wakeup && do_wakeup) {
		spin_lock_irqsave(&i2c_pxa_irqlock, flags);
		i2c_pending = 1;
		spin_unlock_irqrestore(&i2c_pxa_irqlock, flags);
		wake_up_interruptible(&i2c_wait);
	}
	return IRQ_HANDLED;
}

static int i2c_pxa_resource_init(void)
{
	init_waitqueue_head(&i2c_wait);

	if (request_irq(IRQ_I2C, &i2c_pxa_handler, 0, "I2C", 0) < 0) {
		irq = 0;
		if (i2c_debug)
			printk(KERN_INFO "I2C: Failed to register I2C irq %i\n", IRQ_I2C);
		return -ENODEV;
	}
	return 0;
}

static void i2c_pxa_resource_release( void)
{
	if (irq > 0) {
		free_irq(irq,0);
		irq=0;
	}
}

static struct i2c_algo_pxa_data i2c_pxa_data = {
	write_byte:		i2c_pxa_write_byte,
	read_byte:		i2c_pxa_read_byte,

	start:			i2c_pxa_start,
	repeat_start:		i2c_pxa_repeat_start,
	stop:			i2c_pxa_stop,
	abort:			i2c_pxa_abort,

	wait_bus_not_busy:	i2c_pxa_wait_bus_not_busy,
	wait_for_interrupt:	i2c_pxa_wait_for_int,
	transfer:		i2c_pxa_transfer,
	reset:			i2c_pxa_reset,
    slave_recv:		i2c_pxa_slave_recv,

	udelay:			10,
	timeout:		DEF_TIMEOUT,
};

static struct i2c_adapter i2c_pxa_ops = {
	.owner		= THIS_MODULE,
	.id		= I2C_ALGO_PXA,
	.name		= "PXA i2c adapter",
	.algo_data	= &i2c_pxa_data,
	.retries	= 2,
};

extern int i2c_pxa_add_bus(struct i2c_adapter *);
extern int i2c_pxa_del_bus(struct i2c_adapter *);

static int __init i2c_adap_pxa_init(void)
{
	if (i2c_pxa_resource_init() == 0) {
		if (i2c_pxa_add_bus(&i2c_pxa_ops) < 0) {
			i2c_pxa_resource_release();
			printk(KERN_INFO "I2C: Failed to add bus\n");
			return -ENODEV;
		}
	} else {
		return -ENODEV;
	}

	return 0;
}

static void i2c_adap_pxa_exit(void)
{
	i2c_pxa_del_bus(&i2c_pxa_ops);
	i2c_pxa_resource_release();
}

MODULE_DESCRIPTION("PXA2x0 i2c driver");
MODULE_LICENSE("GPL");

module_init(i2c_adap_pxa_init);
module_exit(i2c_adap_pxa_exit);
