#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <asm/uaccess.h>

/* Callback functions for robot_fops (File Operations) */
static int robot_open(struct inode *inode, struct file *filp);

/* Callback functions for brainstem_fops (File Operations) */
static ssize_t brainstem_read (struct file *file, char __user *buf, size_t count,
                            loff_t *offset);
static int brainstem_open(struct inode *inode, struct file *filp);
static ssize_t brainstem_write (struct file *file, const char __user *buf, size_t count,
                             loff_t *offset);
static int brainstem_release(struct inode *inode, struct file *file);

/* Callback functions for brainstem_driver (an i2c_driver) */
static int brainstem_attach_adapter(struct i2c_adapter *adap);
static int brainstem_detach_adapter(struct i2c_adapter *adap);
static int brainstem_detach_client(struct i2c_client *client);
static int brainstem_command(struct i2c_client *client, unsigned int cmd, void *arg);

static struct class_simple *robot_class;

static struct file_operations robot_fops = {
	.owner		= THIS_MODULE,
	.open		= robot_open,
};

static struct file_operations brainstem_fops = {
    .owner		= THIS_MODULE,
    .read		= brainstem_read,
    .llseek		= no_llseek,
    .write		= brainstem_write,
    .open		= brainstem_open,
    .release		= brainstem_release,
};


static struct i2c_driver brainstem_driver = {
    .owner		= THIS_MODULE,
    .name		= "brainstem driver",
    .id			= 0xF1D0, /* I made this up. it's in an ok range */
    .flags		= I2C_DF_NOTIFY,
    .attach_adapter	= brainstem_attach_adapter,
    .detach_adapter	= brainstem_detach_adapter,
    .detach_client	= brainstem_detach_client,
    .command		= brainstem_command,
};

static struct i2c_client brainstem_client_template = {
    .name		= "Brainstem /dev entry",
    .id		= 1,
    .addr		= -1,
    .driver		= &brainstem_driver,
};

/***************************************************************
* Module init and exit functions.  This is where we hook into  *
* the OS.                                                      *      
***************************************************************/
static int __init robot_dev_init(void)
{
	int result;

	result = register_chrdev(42, "robot", &robot_fops);
	if (result)
        {
            printk(KERN_ERR "%s: Driver Initialisation failed\n", __FILE__);
            return result;
        }

        i2c_add_driver(&brainstem_driver);
        
        robot_class = class_simple_create(THIS_MODULE, "robot");
        
         devfs_mk_dir("robot"); 
        class_simple_device_add(robot_class, MKDEV(42,1), NULL, "brainstem");
        devfs_mk_cdev(MKDEV(42,1), S_IFCHR|S_IRUSR|S_IWUSR, "brainstem");

	return 0;
}

static void __exit robot_dev_exit(void)
{
    i2c_del_driver(&brainstem_driver);
    class_simple_destroy(robot_class); 
    devfs_remove("robot");
    unregister_chrdev(42,"robot");
}

/***************************************************************
* Functions needed to implement robot device (Major device     *
* #42)                                                         *
*                                                              *
* 42 is just a number we made up.  It happens to be in a block *
* reserved for experimental/local use, just like this.         *
***************************************************************/
static int robot_open(struct inode *inode, struct file *filp)
{
    switch (iminor(inode)){
        case 1:
            filp->f_op = &brainstem_fops;
            break;
        default:
            return -ENXIO;
    }
    if(filp->f_op && filp->f_op->open)
        return filp->f_op->open(inode, filp);
    return 0;
}

/***************************************************************
* File operation functions for the brainstem device.           *
***************************************************************/
static int brainstem_open(struct inode *inode, struct file *filp)
{
    struct i2c_client *client;
    struct i2c_adapter *adap;

    /* this is bad.  we need to figure out a way to handle */
    /* systems with multiple i2c busses                    */
	/* Fortunately the Gumstix we are using only has one   */
    adap = i2c_get_adapter(0);
    if (!adap)
        return -ENODEV;

    client = kmalloc(sizeof(*client), GFP_KERNEL);
    if (!client) {
        i2c_put_adapter(adap);
        return -ENOMEM;
    }
    memcpy(client, &brainstem_client_template, sizeof(*client));

    client->adapter = adap;
    filp->private_data = client;

    return 0;
}

static ssize_t brainstem_read (struct file *file, char __user *buf, size_t count,
                               loff_t *offset)
{
    char buffer[32];
	int i2csize;
	//int i;
	struct i2c_client *client = (struct i2c_client *)file->private_data;
	
	i2csize = client->adapter->algo->slave_recv(client->adapter, buffer, 32);
	
	/*
	printk("read: ");
	for(i=0;i<i2csize; i++)
	  printk("%d ", buffer[i]);
	printk("\n");
	*/
    if(copy_to_user(buf,buffer,i2csize))
	{
        return 0;
	}
	return i2csize ;
}

/* This is the i2cdev_write function from i2c-dev.c */
static ssize_t brainstem_write (struct file *file, const char __user *buf, size_t count,
                             loff_t *offset)
{
    int ret;
    char *tmp;
	//int i;
    struct i2c_client *client = (struct i2c_client *)file->private_data;

    if (count > 8192)
        count = 8192;

    /* We are expecting a brainstem packet which is at least 3 bytes */
    if (count < 3)
        return 0;
    
    tmp = kmalloc(count,GFP_KERNEL);
    if (tmp==NULL)
        return -ENOMEM;
    if (copy_from_user(tmp,buf,count)) {
        kfree(tmp);
        return -EFAULT;
    }

    client->addr = tmp[0]/2;
    if (count < tmp[1] + 2)
    {
        printk("brainstem: you didn't send enough data for that packet.\n");
        kfree(tmp);
        return -EFAULT;
    }
	/*
	printk("i2c_master_send(%d, \"", client->addr);
	for (i=0;i<tmp[1];i++)
	  printk("%d ", tmp[i+2]);
	printk("\" %d\n", tmp[1]);
	*/
    ret = i2c_master_send(client,&tmp[2],tmp[1]);

    kfree(tmp);

    return ret;
}

static int brainstem_release(struct inode *inode, struct file *file)
{
    struct i2c_client *client = file->private_data;

    i2c_put_adapter(client->adapter);
    kfree(client);
    file->private_data = NULL;

    return 0;
}

/***************************************************************
* Brainstem driver callbacks for being an i2c_device.          *
***************************************************************/
static int brainstem_attach_adapter(struct i2c_adapter *adap)
{
    return 0;
}

static int brainstem_detach_adapter(struct i2c_adapter *adap)
{
    return 0;
}

static int brainstem_detach_client(struct i2c_client *client)
{
    return 0;
}

static int brainstem_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
    return -1;
}

/***************************************************************
* The magic stuff to make this thing a kernel module.          *
***************************************************************/
MODULE_AUTHOR("Brian Davidson <bdavids1@gmu.edu>");
MODULE_DESCRIPTION("robot /dev entries driver");
MODULE_LICENSE("GPL");

module_init(robot_dev_init);
module_exit(robot_dev_exit);
